/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "pch.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "BasicConfigDialog.h"
#include "Application.h"
#include "OperationLog.h"
#include "GlobalState.h"
#include "Conf.h"
#include "ConfigFile.h"
#include "StringUtils.h"
#include "ProxyDescriptor.h"
#include "SymbolicInetAddr.h"
#include "EffectiveFileTimestamps.h"
#include "ArraySize.h"
#include "types.h"
#include <ace/config-lite.h>
#include <boost/lexical_cast.hpp>
#include <wx/textctrl.h>
#include <wx/checkbox.h>
#include <wx/choice.h>
#include <wx/sizer.h>
#include <wx/msgdlg.h>
#include <list>
#include <sstream>
#include <time.h>

using namespace std;

struct BasicConfigDialog::ProxyOption
{
	ProxyDescriptor::ProxyType type;
	wxChar const* label;
};


enum {
	ID_PROXY_TYPE
};


BEGIN_EVENT_TABLE(BasicConfigDialog, wxDialog)
	EVT_CLOSE(onWindowClose)
	EVT_BUTTON(wxID_OK, onApply)
	EVT_CHOICE(ID_PROXY_TYPE, onProxyTypeChanged)
END_EVENT_TABLE()


BasicConfigDialog* BasicConfigDialog::m_spInstance = 0;

BasicConfigDialog::ProxyOption const BasicConfigDialog::m_sProxyOptions[] = {
	{ ProxyDescriptor::INVALID, _T("No forwarding") },
	{ ProxyDescriptor::HTTP, _T("HTTP proxy") },
	{ ProxyDescriptor::SOCKS4, _T("SOCKS4 proxy") },
	{ ProxyDescriptor::SOCKS4A, _T("SOCKS4A proxy") },
	{ ProxyDescriptor::SOCKS5, _T("SOCKS5 proxy") }
};


BasicConfigDialog::BasicConfigDialog()
:	wxDialog(wxGetApp().GetTopWindow(), -1, _T("Basic Configuration"), wxDefaultPosition),
	m_pListenPort(0),
	m_pProxyType(0),
	m_pProxyHost(0),
	m_pProxyPort(0),
	m_pProxyUser(0),
	m_pProxyPass(0)
{
	wxBoxSizer *topsizer = new wxBoxSizer(wxVERTICAL);
	SetSizer(topsizer);
	
	wxBoxSizer *hsz1 = new wxBoxSizer(wxHORIZONTAL);
	topsizer->Add(hsz1, 0, wxALIGN_CENTRE|wxALL, 5 );
	hsz1->Add(new wxStaticText(this, -1, _T("Listen Port")), 0, wxALIGN_CENTRE|wxALL, 5);
	m_pListenPort = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxSize(45, -1));
	hsz1->Add(m_pListenPort, 0, wxALIGN_CENTRE|wxALL, 5);
	
	wxStaticBoxSizer *fwd_sizer = new wxStaticBoxSizer(
		new wxStaticBox(this, -1, _T("Forwarding")),
		wxVERTICAL
	);
	topsizer->Add(fwd_sizer, 0, wxGROW|wxALIGN_CENTER_VERTICAL|wxLEFT|wxRIGHT, 5);
	
	m_pProxyType = new wxChoice(this, ID_PROXY_TYPE);
	fwd_sizer->Add(m_pProxyType, 0, wxGROW|wxALIGN_CENTER_VERTICAL);
	for (int i = 0; i < (int)ARRAY_SIZE(m_sProxyOptions); ++i) {
		m_pProxyType->Append(m_sProxyOptions[i].label);
	}
	
	fwd_sizer->AddSpacer(3);
	
	wxBoxSizer *hsz21 = new wxBoxSizer(wxHORIZONTAL);
	fwd_sizer->Add(hsz21, 0, wxALIGN_CENTRE);
	m_pProxyHost = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxSize(120,-1));
	hsz21->Add(m_pProxyHost, 0, wxALIGN_CENTRE|wxRIGHT, 5);
	hsz21->Add(new wxStaticText(this, -1, _T(":")), 0, wxALIGN_CENTRE);
	m_pProxyPort = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxSize(45,-1));
	hsz21->Add(m_pProxyPort, 0, wxALIGN_CENTRE|wxLEFT, 5);
	
	fwd_sizer->AddSpacer(3);
	
	wxFlexGridSizer* user_pass_sizer = new wxFlexGridSizer(2, 3, 3);
	fwd_sizer->Add(user_pass_sizer, 0, wxGROW);
	user_pass_sizer->AddGrowableCol(1, 1);
	
	user_pass_sizer->Add(
		new wxStaticText(this, -1, _T("Username:")),
		0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL
	);
	m_pProxyUser = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxSize(100,-1));
	user_pass_sizer->Add(m_pProxyUser, 0, wxGROW);
	
	user_pass_sizer->Add(
		new wxStaticText(this, -1, _T("Password:")),
		0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL
	);
	m_pProxyPass = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, wxSize(100,-1));
	user_pass_sizer->Add(m_pProxyPass, 0, wxGROW);
	
	m_pTrayAnimation = new wxCheckBox(this, -1, _T("Tray icon animation"));
	topsizer->Add(m_pTrayAnimation, 0, wxLEFT|wxRIGHT|wxTOP, 5);
	
	wxBoxSizer *hsz3 = new wxBoxSizer(wxHORIZONTAL);
	topsizer->Add(hsz3, 0, wxALIGN_CENTRE|wxALL, 5);
	hsz3->Add(new wxButton(this, wxID_OK, _T("OK")), 0, wxALIGN_CENTRE|wxALL, 5);
	wxButton* cancel_btn = new wxButton(this, wxID_CANCEL, _T("Cancel"));
	hsz3->Add(cancel_btn, 0, wxALIGN_CENTRE|wxALL, 5);
	cancel_btn->SetDefault();
	
	topsizer->SetSizeHints(this);
	
	setProxyType(ProxyDescriptor::INVALID);
	loadConfig();
}

BasicConfigDialog::~BasicConfigDialog()
{
}

void
BasicConfigDialog::show()
{
	if (!m_spInstance) {
		m_spInstance = new BasicConfigDialog;
		m_spInstance->Show();
		m_spInstance->m_pListenPort->SetSelection(0, 0);
		// Why the hell it gets selected?
		// And why unselecting it only works after Show()?
	} else {
		m_spInstance->Show();
		m_spInstance->Raise();
	}
}

void
BasicConfigDialog::onWindowClose(wxCloseEvent& evt)
{
	if (m_spInstance == this) {
		m_spInstance = 0;
	}
	Destroy();
}

void
BasicConfigDialog::onProxyTypeChanged(wxCommandEvent& evt)
{
	onProxyTypeChanged(m_sProxyOptions[evt.GetSelection()].type);
}

void
BasicConfigDialog::onProxyTypeChanged(ProxyDescriptor::ProxyType type)
{
	switch (type) {
		case ProxyDescriptor::INVALID:
		m_pProxyHost->Enable(false);
		m_pProxyPort->Enable(false);
		m_pProxyUser->Enable(false);
		m_pProxyPass->Enable(false);
		break;
		case ProxyDescriptor::HTTP:
		m_pProxyHost->Enable(true);
		m_pProxyPort->Enable(true);
		m_pProxyUser->Enable(false);
		m_pProxyPass->Enable(false);
		break;
		case ProxyDescriptor::SOCKS4:
		case ProxyDescriptor::SOCKS4A:
		m_pProxyHost->Enable(true);
		m_pProxyPort->Enable(true);
		m_pProxyUser->Enable(true);
		m_pProxyPass->Enable(false);
		break;
		case ProxyDescriptor::SOCKS5:
		m_pProxyHost->Enable(true);
		m_pProxyPort->Enable(true);
		m_pProxyUser->Enable(true);
		m_pProxyPass->Enable(true);
		break;
	}
}

ProxyDescriptor::ProxyType
BasicConfigDialog::getProxyType() const
{
	return m_sProxyOptions[m_pProxyType->GetSelection()].type;
}
	
void
BasicConfigDialog::setProxyType(ProxyDescriptor::ProxyType type)
{
	for (int i = 0; i < (int)ARRAY_SIZE(m_sProxyOptions); ++i) {
		if (m_sProxyOptions[i].type == type) {
			m_pProxyType->SetSelection(i);
			onProxyTypeChanged(type);
			break;
		}
	}
}

void
BasicConfigDialog::loadConfig()
{
	Config config(GlobalState::ReadAccessor()->config());
	list<SymbolicInetAddr> listen_addrs(config.getListenAddresses());
	if (listen_addrs.size() > 1) {
		setCurrentListenPort(wxEmptyString),
		showWarning(
			_T("Your current config has multiple listen addresses.\n")
			_T("You should be using the Advanced Configuration dialog.")
		);
	} else if (listen_addrs.size() == 1) {
		wxString val;
		val << listen_addrs.front().getPort();
		setCurrentListenPort(val);
	} else {
		setCurrentListenPort(wxEmptyString);
	}
	m_pTrayAnimation->SetValue(config.isTrayAnimationEnabled());
	
	ProxyDescriptor pdesc(config.getNextHopProxy());
	setProxyType(config.isNextHopProxyEnabled() ? pdesc.getType() : pdesc.INVALID);
	
	m_pProxyHost->SetValue(pdesc.getAddr().getHost().c_str());
	if (pdesc.getAddr().getPort() <= 0) {
		m_pProxyPort->SetValue(wxEmptyString);
	} else {
		wxString val;
		val << pdesc.getAddr().getPort();
		m_pProxyPort->SetValue(val);
	}
	m_pProxyUser->SetValue(pdesc.getUserName().c_str());
	m_pProxyPass->SetValue(pdesc.getPassword().c_str());
}

void
BasicConfigDialog::setCurrentListenPort(wxString const& port)
{
	m_curListenPort = port;
	m_pListenPort->SetValue(port);
}

void
BasicConfigDialog::onApply(wxCommandEvent& evt)
{
	bool const listen_port_changed = (m_pListenPort->GetValue() != m_curListenPort);
	uint16_t listen_port = 0;
	if (listen_port_changed) {
		try {
			string port = StringUtils::trimStr(m_pListenPort->GetValue().mb_str());
			listen_port = boost::lexical_cast<uint16_t>(port);
		} catch (boost::bad_lexical_cast& e) {}
		if (listen_port == 0) {
			showError(_T("Listen port must be an integer in range of [1, 65535]"));
			m_pListenPort->SetFocus();
			return;
		}
	}
	
	ProxyDescriptor::ProxyType const proxy_type = getProxyType();
	bool const forwarding = (proxy_type != ProxyDescriptor::INVALID);
	
	string proxy_host = StringUtils::trimStr(m_pProxyHost->GetValue().mb_str());	
	if (forwarding && proxy_host.empty()) {
		showError(_T("Remote proxy host is not given"));
		m_pProxyHost->SetFocus();
		return;
	}
	
	uint16_t proxy_port = 0;
	try {
		string port = StringUtils::trimStr(m_pProxyPort->GetValue().mb_str());
		proxy_port = boost::lexical_cast<uint16_t>(port);
	} catch (boost::bad_lexical_cast& e) {}
	if (forwarding && proxy_port == 0) {
		showError(_T("Remote proxy port must be an integer in range of [1, 65535]"));
		m_pProxyPort->SetFocus();
		return;
	}
	
	string proxy_user;
	string proxy_pass;
	if (proxy_type != ProxyDescriptor::HTTP || forwarding == false) {
		proxy_user = m_pProxyUser->GetValue().mb_str();
	}
	if (proxy_type == ProxyDescriptor::SOCKS5 || forwarding == false) {
		proxy_pass = m_pProxyPass->GetValue().mb_str();
	}
	
	Config const old_config(GlobalState::ReadAccessor()->config());
	Config new_config(old_config);
	if (listen_port_changed) {
		new_config.setListenAddresses(
			list<SymbolicInetAddr>(1, SymbolicInetAddr("127.0.0.1", listen_port))
		);
	}
	new_config.setTrayAnimationEnabled(m_pTrayAnimation->GetValue());
	new_config.setNextHopProxyEnabled(forwarding);
	
	SymbolicInetAddr proxy_addr(proxy_host, -1);
	if (proxy_port > 0) {
		proxy_addr.setPort(proxy_port);
	}
	ProxyDescriptor proxy(old_config.getNextHopProxy());
	proxy.setType(proxy_type);
	proxy.setAddr(proxy_addr);
	proxy.setUserName(proxy_user);
	proxy.setPassword(proxy_pass);
	new_config.setNextHopProxy(proxy);
	
	Application& app = wxGetApp();
	Log* log = OperationLog::instance();
	log->appendRecord(_T("Applying new config ... "));
	size_t num_records = log->getNumRecords();
	
	if (!app.applyConfig(new_config)) {
		restoreOldConfig(old_config);
		app.showLogDialog();
		return;
	}
	
	ConfigFile new_config_file(app.configFile());
	new_config_file.updateWith(new_config);
	ostringstream content;
	content << new_config_file;
	
	time_t mtime = 0;
	if (!app.writeFile(app.getConfigFileName(), content.str(), &mtime)) {
		restoreOldConfig(old_config);
		app.showLogDialog();
		return;
	}
	
	app.configFile().swap(new_config_file);
	EffectiveFileTimestamps::config = mtime;
	
	if (num_records == log->getNumRecords()) {
		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
	}
	
	Close();
}

bool
BasicConfigDialog::restoreOldConfig(Config const& config)
{
	Application& app = wxGetApp();
	Log* log = OperationLog::instance();
	log->appendRecord(_T("Restoring the old config ... "));
	size_t num_records = log->getNumRecords();
	
	if (!app.applyConfig(config)) {
		return false;
	}
	
	if (num_records == log->getNumRecords()) {
		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
	}
	return true;
}

void
BasicConfigDialog::showError(wxString const& msg)
{
	wxMessageDialog* dialog = new wxMessageDialog(
		this, msg, _T("Error"), wxOK|wxICON_ERROR
	);
	dialog->ShowModal();
}

void
BasicConfigDialog::showWarning(wxString const& msg)
{
	wxMessageDialog* dialog = new wxMessageDialog(
		this, msg, _T("Warning"), wxOK|wxICON_EXCLAMATION
	);
	dialog->ShowModal();
}
